# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import zlib

from typing import Dict, Union
from enum import Enum

from dataclasses import dataclass
from dataclasses_json import dataclass_json


class QuantumNoiseModelSerializationFormat(Enum):
    UNKOWN_SERIALIZATION_FORMAT = 0
    QISKIT_AER_JSON_V1 = 1
    QISKIT_AER_ZLIB_JSON_V1 = 2


@dataclass_json
@dataclass
class QuantumNoiseModel:
    serialization_format: QuantumNoiseModelSerializationFormat
    serialization: bytes

    @classmethod
    def from_dict(cls, data: Union[Dict, str]) -> "QuantumNoiseModel":
        return QuantumNoiseModel.schema().load(data)

    def to_dict(self) -> Dict:
        return QuantumNoiseModel.schema().dump(self)

    @classmethod
    def from_json(cls, st: str) -> "QuantumNoiseModel":
        data = json.loads(data) if isinstance(data, str) else data
        return cls.from_dict(data)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_qiskit_aer_noise_model(
        noise_model: "qiskit_aer.NoiseModel",
    ) -> "QuantumNoiseModel":
        try:
            import numpy as np
        except ImportError:
            raise Exception("Numpy is not installed")

        try:
            from qiskit_aer.noise import NoiseModel
        except ImportError:
            raise Exception("Qiskit Aer is not installed")

        def _encode_numpy_complex(obj):
            """
            Recursively traverses a structure and converts numpy arrays and
            complex numbers into a JSON-serializable format.
            """
            if isinstance(obj, np.ndarray):
                return {
                    "__ndarray__": True,
                    "data": _encode_numpy_complex(
                        obj.tolist()
                    ),  # Recursively encode data
                    "dtype": obj.dtype.name,
                    "shape": obj.shape,
                }
            elif isinstance(obj, (complex, np.complex128)):
                return {"__complex__": True, "real": obj.real, "imag": obj.imag}
            elif isinstance(obj, dict):
                return {key: _encode_numpy_complex(value) for key, value in obj.items()}
            elif isinstance(obj, (list, tuple)):
                return [_encode_numpy_complex(item) for item in obj]
            else:
                return obj

        noise_model_dict = _encode_numpy_complex(noise_model.to_dict(False))
        return QuantumNoiseModel(
            serialization_format=QuantumNoiseModelSerializationFormat.QISKIT_AER_ZLIB_JSON_V1,
            serialization=zlib.compress(json.dumps(noise_model_dict).encode()),
            frozenset=True,
        )

    def to_qiskit_aer_noise_model(self) -> "qiskit_aer.NoiseModel":
        try:
            import numpy as np
        except ImportError:
            raise Exception("Numpy is not installed")

        try:
            from qiskit_aer.noise import NoiseModel
        except ImportError:
            raise Exception("Qiskit Aer is not installed")

        def _custom_decode_numpy_and_complex(obj):
            """
            Recursively traverses a structure and converts dictionary representations
            back into numpy arrays and complex numbers.
            """
            if isinstance(obj, dict):
                if obj.get("__ndarray__", False):
                    # The data has been recursively decoded, so we can build the array
                    return np.array(
                        _custom_decode_numpy_and_complex(obj["data"]),
                        dtype=obj["dtype"],
                    ).reshape(obj["shape"])
                elif obj.get("__complex__", False):
                    return complex(obj["real"], obj["imag"])
                else:
                    return {
                        key: _custom_decode_numpy_and_complex(value)
                        for key, value in obj.items()
                    }
            elif isinstance(obj, list):
                return [_custom_decode_numpy_and_complex(item) for item in obj]
            else:
                return obj

        def _json_deserialization_noise_model(noise_model_str: bytes):
            noise_model_dict = json.loads(noise_model_str)
            noise_model = NoiseModel.from_dict(noise_model_dict)

            return noise_model

        def _zlib_json_deserialization_noise_model(noise_model_raw: bytes):
            """
            Attempt to decode the noise model dictionary using the custom decoder, and if it fails, try to use the default decoder.
            This is a workaround for a bug in the Qiskit NoiseModel.from_dict method.
            """
            noise_model_str = zlib.decompress(noise_model_raw)
            noise_model_dict = json.loads(noise_model_str)
            try:
                noise_model_dict_custom_decoded = _custom_decode_numpy_and_complex(
                    noise_model_dict
                )
            except:
                return NoiseModel.from_dict(noise_model_dict)
            return NoiseModel.from_dict(noise_model_dict_custom_decoded)

        match = {
            QuantumNoiseModelSerializationFormat.QISKIT_AER_JSON_V1: _json_deserialization_noise_model,
            QuantumNoiseModelSerializationFormat.QISKIT_AER_ZLIB_JSON_V1: _zlib_json_deserialization_noise_model,
        }

        try:
            return match[self.serialization_format](self.serialization)
        except:
            raise Exception(
                "unsupported serialization format:", self.serialization_format
            )



# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.from enum import Enum
import json

from enum import Enum
from typing import Dict, Union

from dataclasses import dataclass
from dataclasses_json import dataclass_json


class QuantumProgramSerializationFormat(Enum):
    UNKOWN_SERIALIZATION_FORMAT = 0
    QASM_V1 = 1
    QASM_V2 = 2
    QASM_V3 = 3
    QIR_V1 = 4
    CIRQ_CIRCUIT_JSON_V1 = 5
    PERCEVAL_CIRCUIT_JSON_V1 = 6
    PULSER_SEQUENCE_JSON_V1 = 7


@dataclass_json
@dataclass
class QuantumProgram:
    serialization_format: QuantumProgramSerializationFormat
    serialization: str

    @classmethod
    def from_dict(cls, data: Union[Dict, str]) -> "QuantumProgram":
        return QuantumProgram.schema().load(data)

    def to_dict(self) -> Dict:
        return QuantumProgram.schema().dump(self)

    @classmethod
    def from_json(cls, str: str) -> "QuantumProgram":
        data = json.loads(data) if isinstance(data, str) else data
        return cls.from_dict(data)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_qiskit_circuit(
        cls,
        qiskit_circuit: "qiskit.QuantumCircuit",
        dest_format: QuantumProgramSerializationFormat = QuantumProgramSerializationFormat.QASM_V3,
    ) -> "QuantumProgram":
        try:
            from qiskit import qasm3, qasm2
        except ImportError:
            raise Exception("Qiskit is not installed")

        match = {
            QuantumProgramSerializationFormat.QASM_V2: lambda c: qasm2.dumps(c),
            QuantumProgramSerializationFormat.QASM_V3: lambda c: qasm3.dumps(c),
        }

        try:
            return cls(
                serialization_format=dest_format,
                serialization=match[dest_format](qiskit_circuit),
                frozenset=True,
            )
        except:
            raise

    def to_qiskit_circuit(self) -> "qiskit.QuantumCircuit":
        try:
            from qiskit import qasm3, qasm2, QuantumCircuit
        except ImportError:
            raise Exception("Qiskit is not installed")

        match = {
            QuantumProgramSerializationFormat.QASM_V1: lambda c: QuantumCircuit.from_qasm_str(
                c
            ),
            QuantumProgramSerializationFormat.QASM_V2: lambda c: qasm2.loads(c),
            QuantumProgramSerializationFormat.QASM_V3: lambda c: qasm3.loads(c),
        }

        try:
            return match[self.serialization_format](self.serialization)
        except:
            raise Exception(
                "unsupported serialization format:", self.serialization_format
            )

    @classmethod
    def from_cirq_circuit(
        cls,
        cirq_circuit: "cirq.AbstractCircuit",
        dest_format: QuantumProgramSerializationFormat = QuantumProgramSerializationFormat.QASM_V3,
    ) -> "QuantumProgram":
        try:
            import cirq
        except ImportError:
            raise Exception("Cirq is not installed")

        match = {
            QuantumProgramSerializationFormat.QASM_V2: lambda c: cirq.qasm(c),
            QuantumProgramSerializationFormat.QASM_V3: lambda c: cirq.qasm(
                c, args=cirq.QasmArgs(version="3.0")
            ),
            QuantumProgramSerializationFormat.CIRQ_CIRCUIT_JSON_V1: lambda c: cirq.to_json(
                c
            ),
        }

        try:
            return cls(
                serialization_format=dest_format,
                serialization=match[dest_format](cirq_circuit),
                frozenset=True,
            )
        except:
            raise Exception("unsupported unserialization:", dest_format)

    def to_cirq_circuit(self) -> "cirq.Circuit":
        try:
            import cirq
        except ImportError:
            raise Exception("Cirq is not installed")

        if self.serialization_format in [
            QuantumProgramSerializationFormat.QASM_V1,
            QuantumProgramSerializationFormat.QASM_V2,
        ]:
            from cirq.contrib.qasm_import import circuit_from_qasm

            return circuit_from_qasm(self.serialization)

        if self.serialization_format in [
            QuantumProgramSerializationFormat.CIRQ_CIRCUIT_JSON_V1,
        ]:
            from cirq import read_json

            return read_json(json_text=self.serialization)

        raise Exception("unsupported serialization format:", self.serialization_format)



# Copyright 2025 Scaleway, Aqora, Quantum Commons
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from typing import Union, Dict, Tuple
from enum import Enum

from dataclasses import dataclass
from dataclasses_json import dataclass_json


class QuantumProgramResultSerializationFormat(Enum):
    UNKOWN_SERIALIZATION_FORMAT = 0
    CIRQ_RESULT_JSON_V1 = 1
    QISKIT_RESULT_JSON_V1 = 2


@dataclass_json
@dataclass
class QuantumProgramResult:
    serialization_format: QuantumProgramResultSerializationFormat
    serialization: str

    @classmethod
    def from_dict(cls, data: Union[Dict, str]) -> "QuantumProgramResult":
        return QuantumProgramResult.schema().load(data)

    def to_dict(self) -> Dict:
        return QuantumProgramResult.schema().dump(self)

    @classmethod
    def from_json(cls, str: str) -> "QuantumProgramResult":
        data = json.loads(data) if isinstance(data, str) else data
        return cls.from_dict(data)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_qiskit_result(
        cls, qiskit_result: "qiskit.result.Result"
    ) -> "QuantumProgramResult":
        try:
            from qiskit.result import Result
        except ImportError:
            raise Exception("Qiskit is not installed")

        serialization = json.dumps(qiskit_result.to_dict())

        return cls(
            serialization_format=QuantumProgramResultSerializationFormat.QISKIT_RESULT_JSON_V1,
            serialization=serialization,
            frozenset=True,
        )

    @classmethod
    def from_qiskit_result_dict(
        cls, qiskit_result_dict: Union[str, Dict]
    ) -> "QuantumProgramResult":
        if isinstance(qiskit_result_dict, str):
            qiskit_result_dict = json.loads(
                qiskit_result_dict
            )  # Ensure serialization is not ill-formatted

        serialization = json.dumps(qiskit_result_dict)

        return cls(
            serialization_format=QuantumProgramResultSerializationFormat.QISKIT_RESULT_JSON_V1,
            serialization=serialization,
            frozenset=True,
        )

    def to_qiskit_result(self, **kwargs) -> "qiskit.result.Result":
        try:
            from qiskit.result import Result
            from qiskit.result import Result
            from qiskit.result.models import ExperimentResult, ExperimentResultData

        except ImportError:
            raise Exception("Qiskit is not installed")

        result_dict = json.loads(self.serialization)

        if (
            self.serialization_format
            == QuantumProgramResultSerializationFormat.QISKIT_RESULT_JSON_V1
        ):
            return Result.from_dict(
                {
                    "results": result_dict["results"],
                    "success": result_dict["success"],
                    "header": result_dict.get("header"),
                    "metadata": result_dict.get("metadata"),
                }.update(kwargs)
            )
        elif (
            self.serialization_format
            == QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1
        ):
            from cirq.study import ResultDict

            def __make_hex_from_result_array(result: Tuple):
                str_value = "".join(map(str, result))
                integer_value = int(str_value, 2)

                return hex(integer_value)

            def __make_expresult_from_cirq_result(
                cirq_result: ResultDict,
            ) -> ExperimentResult:
                hist = dict(
                    cirq_result.multi_measurement_histogram(
                        keys=cirq_result.measurements.keys()
                    )
                )

                return ExperimentResult(
                    shots=cirq_result.repetitions,
                    success=True,
                    data=ExperimentResultData(
                        counts={
                            __make_hex_from_result_array(key): value
                            for key, value in hist.items()
                        },
                    ),
                )

            result_dict = json.loads(self.serialization)
            cirq_result = ResultDict._from_json_dict_(**result_dict)

            return Result.from_dict(
                {
                    "results": [__make_expresult_from_cirq_result(cirq_result)],
                }.update(kwargs)
            )
        else:
            raise Exception(
                "unsupported serialization format:", self.serialization_format
            )

    @classmethod
    def from_cirq_result_dict(
        cls, cirq_result_dict: Union[str, Dict]
    ) -> "QuantumProgramResult":
        if isinstance(cirq_result_dict, str):
            cirq_result_dict = json.loads(
                cirq_result_dict
            )  # Ensure serialization is not ill-formatted

        serialization = json.dumps(cirq_result_dict)

        return cls(
            serialization_format=QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1,
            serialization=serialization,
            frozenset=True,
        )

    @classmethod
    def from_cirq_result(
        cls, cirq_result: "cirq.Result", **kwargs
    ) -> "QuantumProgramResult":
        try:
            import cirq
        except ImportError:
            raise Exception("Cirq is not installed")

        serialization = json.dumps(cirq_result._json_dict_().update(kwargs))

        return cls(
            serialization_format=QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1,
            serialization=serialization,
            frozenset=True,
        )

    def to_cirq_result(self, **kwargs) -> "cirq.Result":
        try:
            from cirq import ResultDict
        except ImportError:
            raise Exception("Cirq is not installed")

        if (
            self.serialization_format
            != QuantumProgramResultSerializationFormat.CIRQ_RESULT_JSON_V1
        ):
            raise Exception(
                "unsupported serialization format:", self.serialization_format
            )

        result_dict = json.loads(self.serialization)
        cirq_result = ResultDict._from_json_dict_(**result_dict)

        return cirq_result
